@@ -41,25 +41,19 @@ abstract type LibuvStream <: IO end
 
 # Redirectable = Union{IO, FileRedirect, Libc.RawFD} (not exported)
 
-function stream_wait(x, c...) # for x::LibuvObject
-    preserve_handle(x)
-    try
-        return wait(c...)
-    finally
-        unpreserve_handle(x)
-    end
-end
-
 bytesavailable(s::LibuvStream) = bytesavailable(s.buffer)
 
 function eof(s::LibuvStream)
-    if isopen(s) # fast path
-        bytesavailable(s) > 0 && return false
-    else
-        return bytesavailable(s) <= 0
-    end
-    wait_readnb(s,1)
-    return !isopen(s) && bytesavailable(s) <= 0
+    bytesavailable(s) > 0 && return false
+    wait_readnb(s, 1)
+    # This function is race-y if used from multiple threads, but we guarantee
+    # it to never return false until the stream is definitively exhausted
+    # and that we won't return true if there's a readerror pending (it'll instead get thrown).
+    # This requires some careful ordering here (TODO: atomic loads)
+    bytesavailable(s) > 0 && return false
+    open = isopen(s) # must preceed readerror check
+    s.readerror === nothing || throw(s.readerror)
+    return !open
 end
 
 # Limit our default maximum read and buffer size,
@@ -81,7 +75,7 @@ const StatusOpen        = 3 # handle is usable
 const StatusActive      = 4 # handle is listening for read/write/connect events
 const StatusClosing     = 5 # handle is closing / being closed
 const StatusClosed      = 6 # handle is closed
-const StatusEOF         = 7 # handle is a TTY that has seen an EOF event
+const StatusEOF         = 7 # handle is a TTY that has seen an EOF event (pretends to be closed until reseteof is called)
 const StatusPaused      = 8 # handle is Active, but not consuming events, and will transition to Open if it receives an event
 function uv_status_string(x)
     s = x.status
@@ -118,19 +112,17 @@ mutable struct PipeEndpoint <: LibuvStream
     handle::Ptr{Cvoid}
     status::Int
     buffer::IOBuffer
-    readnotify::Condition
-    connectnotify::Condition
-    closenotify::Condition
+    cond::ThreadSynchronizer
+    readerror::Any
     sendbuf::Union{IOBuffer, Nothing}
-    lock::ReentrantLock
+    lock::ReentrantLock # advisory lock
     throttle::Int
     function PipeEndpoint(handle::Ptr{Cvoid}, status)
         p = new(handle,
                 status,
                 PipeBuffer(),
-                Condition(),
-                Condition(),
-                Condition(),
+                ThreadSynchronizer(),
+                nothing,
                 nothing,
                 ReentrantLock(),
                 DEFAULT_READ_BUFFER_SZ)
@@ -142,17 +134,21 @@ end
 
 function PipeEndpoint()
     pipe = PipeEndpoint(Libc.malloc(_sizeof_uv_named_pipe), StatusUninit)
+    iolock_begin()
     err = ccall(:uv_pipe_init, Cint, (Ptr{Cvoid}, Ptr{Cvoid}, Cint), eventloop(), pipe.handle, 0)
     uv_error("failed to create pipe endpoint", err)
     pipe.status = StatusInit
+    iolock_end()
     return pipe
 end
 
 function PipeEndpoint(fd::OS_HANDLE)
     pipe = PipeEndpoint()
+    iolock_begin()
     err = ccall(:uv_pipe_open, Int32, (Ptr{Cvoid}, OS_HANDLE), pipe.handle, fd)
     uv_error("pipe_open", err)
     pipe.status = StatusOpen
+    iolock_end()
     return pipe
 end
 if OS_HANDLE != RawFD
@@ -164,10 +160,10 @@ mutable struct TTY <: LibuvStream
     handle::Ptr{Cvoid}
     status::Int
     buffer::IOBuffer
-    readnotify::Condition
-    closenotify::Condition
+    cond::ThreadSynchronizer
+    readerror::Any
     sendbuf::Union{IOBuffer, Nothing}
-    lock::ReentrantLock
+    lock::ReentrantLock # advisory lock
     throttle::Int
     @static if Sys.iswindows(); ispty::Bool; end
     function TTY(handle::Ptr{Cvoid}, status)
@@ -175,8 +171,8 @@ mutable struct TTY <: LibuvStream
             handle,
             status,
             PipeBuffer(),
-            Condition(),
-            Condition(),
+            ThreadSynchronizer(),
+            nothing,
             nothing,
             ReentrantLock(),
             DEFAULT_READ_BUFFER_SZ)
@@ -191,10 +187,12 @@ end
 
 function TTY(fd::OS_HANDLE)
     tty = TTY(Libc.malloc(_sizeof_uv_tty), StatusUninit)
+    iolock_begin()
     err = ccall(:uv_tty_init, Int32, (Ptr{Cvoid}, Ptr{Cvoid}, OS_HANDLE, Int32),
         eventloop(), tty.handle, fd, 0)
     uv_error("TTY", err)
     tty.status = StatusOpen
+    iolock_end()
     return tty
 end
 if OS_HANDLE != RawFD
@@ -230,7 +228,9 @@ rawhandle(stream::LibuvStream) = stream.handle
 unsafe_convert(::Type{Ptr{Cvoid}}, s::Union{LibuvStream, LibuvServer}) = s.handle
 
 function init_stdio(handle::Ptr{Cvoid})
+    iolock_begin()
     t = ccall(:jl_uv_handle_type, Int32, (Ptr{Cvoid},), handle)
+    local io
     if t == UV_FILE
         fd = ccall(:jl_uv_file_handle, OS_HANDLE, (Ptr{Cvoid},), handle)
         # TODO: Replace ios.c file with libuv fs?
@@ -240,17 +240,19 @@ function init_stdio(handle::Ptr{Cvoid})
             fd = ccall(:_open_osfhandle, RawFD, (WindowsRawSocket, Int32), fd, 0)
         end
         # TODO: Get fdio to work natively with file descriptors instead of integers
-        return fdio(cconvert(Cint, fd))
+        io = fdio(cconvert(Cint, fd))
     elseif t == UV_TTY
-        return TTY(handle, StatusOpen)
+        io = TTY(handle, StatusOpen)
     elseif t == UV_TCP
         Sockets = require(PkgId(UUID((0x6462fe0b_24de_5631, 0x8697_dd941f90decc)), "Sockets"))
-        return Sockets.TCPSocket(handle, StatusOpen)
+        io = Sockets.TCPSocket(handle, StatusOpen)
     elseif t == UV_NAMED_PIPE
-        return PipeEndpoint(handle, StatusOpen)
+        io = PipeEndpoint(handle, StatusOpen)
     else
         throw(ArgumentError("invalid stdio type: $t"))
     end
+    iolock_end()
+    return io
 end
 
 """
@@ -266,35 +268,38 @@ of the original handle.
     other part of the system.
 """
 function open(h::OS_HANDLE)
+    iolock_begin()
     t = ccall(:uv_guess_handle, Cint, (OS_HANDLE,), h)
+    local io
     if t == UV_FILE
         @static if Sys.iswindows()
             # TODO: Get ios.c to understand native handles
             h = ccall(:_open_osfhandle, RawFD, (WindowsRawSocket, Int32), h, 0)
         end
         # TODO: Get fdio to work natively with file descriptors instead of integers
-        return fdio(cconvert(Cint, h))
+        io = fdio(cconvert(Cint, h))
     elseif t == UV_TTY
-        return TTY(h)
+        io = TTY(h)
     elseif t == UV_TCP
         Sockets = require(PkgId(UUID((0x6462fe0b_24de_5631, 0x8697_dd941f90decc)), "Sockets"))
-        return Sockets.TCPSocket(h)
+        io = Sockets.TCPSocket(h)
     elseif t == UV_NAMED_PIPE
-        pipe = PipeEndpoint(h)
+        io = PipeEndpoint(h)
         @static if Sys.iswindows()
-            if ccall(:jl_ispty, Cint, (Ptr{Cvoid},), pipe.handle) != 0
+            if ccall(:jl_ispty, Cint, (Ptr{Cvoid},), io.handle) != 0
                 # replace the Julia `PipeEndpoint` type with a `TTY` type,
                 # if we detect that this is a cygwin pty object
-                pipe_handle, pipe_status = pipe.handle, pipe.status
-                pipe.status = StatusClosed
-                pipe.handle = C_NULL
-                return TTY(pipe_handle, pipe_status)
+                pipe_handle, pipe_status = io.handle, pipe.status
+                io.status = StatusClosed
+                io.handle = C_NULL
+                io = TTY(pipe_handle, pipe_status)
             end
         end
-        return pipe
     else
         throw(ArgumentError("invalid stdio type: $t"))
     end
+    iolock_end()
+    return io
 end
 
 if OS_HANDLE != RawFD
@@ -315,7 +320,7 @@ function isopen(x::Union{LibuvStream, LibuvServer})
     if x.status == StatusUninit || x.status == StatusInit
         throw(ArgumentError("$x is not initialized"))
     end
-    x.status != StatusClosed && x.status != StatusEOF
+    return x.status != StatusClosed && x.status != StatusEOF
 end
 
 function check_open(x::Union{LibuvStream, LibuvServer})
@@ -324,84 +329,85 @@ function check_open(x::Union{LibuvStream, LibuvServer})
     end
 end
 
-function wait_connected(x::Union{LibuvStream, LibuvServer})
-    check_open(x)
-    while x.status == StatusConnecting
-        stream_wait(x, x.connectnotify)
-        check_open(x)
-    end
-end
-
-function wait_readbyte(x::LibuvStream, c::UInt8)
-    if isopen(x) # fast path
-        occursin(c, x.buffer) && return
-    else
-        return
-    end
-    preserve_handle(x)
-    try
-        while isopen(x) && !occursin(c, x.buffer)
-            start_reading(x) # ensure we are reading
-            wait(x.readnotify)
-        end
-    finally
-        if isempty(x.readnotify.waitq)
-            stop_reading(x) # stop reading iff there are currently no other read clients of the stream
-        end
-        unpreserve_handle(x)
-    end
-    nothing
-end
-
 function wait_readnb(x::LibuvStream, nb::Int)
-    if isopen(x) # fast path
-        bytesavailable(x.buffer) >= nb && return
-    else
-        return
-    end
+    # fast path before iolock acquire
+    bytesavailable(x.buffer) >= nb && return
+    open = isopen(x) # must preceed readerror check
+    x.readerror === nothing || throw(x.readerror)
+    open || return
+    iolock_begin()
+    # repeat fast path after iolock acquire, before other expensive work
+    bytesavailable(x.buffer) >= nb && (iolock_end(); return)
+    open = isopen(x)
+    x.readerror === nothing || throw(x.readerror)
+    open || (iolock_end(); return)
+    # now do the "real" work
     oldthrottle = x.throttle
     preserve_handle(x)
+    lock(x.cond)
     try
-        while isopen(x) && bytesavailable(x.buffer) < nb
+        while bytesavailable(x.buffer) < nb
+            x.readerror === nothing || throw(x.readerror)
+            isopen(x) || break
             x.throttle = max(nb, x.throttle)
             start_reading(x) # ensure we are reading
-            wait(x.readnotify)
+            iolock_end()
+            wait(x.cond)
+            unlock(x.cond)
+            iolock_begin()
+            lock(x.cond)
         end
     finally
-        if isempty(x.readnotify.waitq)
+        if isempty(x.cond)
             stop_reading(x) # stop reading iff there are currently no other read clients of the stream
         end
         if oldthrottle <= x.throttle <= nb
+            # if we're interleaving readers, we might not get back to the "original" throttle
+            # but we consider that an acceptable "risk", since we can't be quite sure what the intended value is now
             x.throttle = oldthrottle
         end
         unpreserve_handle(x)
+        unlock(x.cond)
     end
+    iolock_end()
     nothing
 end
 
 function wait_close(x::Union{LibuvStream, LibuvServer})
-    if isopen(x)
-        stream_wait(x, x.closenotify)
+    preserve_handle(x)
+    lock(x.cond)
+    try
+        while isopen(x)
+            wait(x.cond)
+        end
+    finally
+        unlock(x.cond)
+        unpreserve_handle(x)
     end
     nothing
 end
 
 function close(stream::Union{LibuvStream, LibuvServer})
+    iolock_begin()
+    should_wait = false
     if stream.status == StatusInit
         ccall(:jl_forceclose_uv, Cvoid, (Ptr{Cvoid},), stream.handle)
-    elseif isopen(stream)
+        stream.status = StatusClosing
+    elseif isopen(stream) || stream.status == StatusEOF
+        should_wait = uv_handle_data(stream) != C_NULL
         if stream.status != StatusClosing
             ccall(:jl_close_uv, Cvoid, (Ptr{Cvoid},), stream.handle)
             stream.status = StatusClosing
         end
-        if uv_handle_data(stream) != C_NULL
-            stream_wait(stream, stream.closenotify)
-        end
     end
+    iolock_end()
+    should_wait && wait_close(stream)
     nothing
 end
 
 function uvfinalize(uv::Union{LibuvStream, LibuvServer})
+    uv.handle == C_NULL && return
+    iolock_begin()
     if uv.handle != C_NULL
         disassociate_julia_struct(uv.handle) # not going to call the usual close hooks
         if uv.status != StatusUninit
@@ -412,6 +418,7 @@ function uvfinalize(uv::Union{LibuvStream, LibuvServer})
         uv.status = StatusClosed
         uv.handle = C_NULL
     end
+    iolock_end()
     nothing
 end
 
@@ -467,9 +474,11 @@ function displaysize(io::TTY)
 
     s1 = Ref{Int32}(0)
     s2 = Ref{Int32}(0)
+    iolock_begin()
     Base.uv_error("size (TTY)", ccall(:uv_tty_get_winsize,
                                       Int32, (Ptr{Cvoid}, Ptr{Int32}, Ptr{Int32}),
                                       io, s1, s2) != 0)
+    iolock_end()
     w, h = s1[], s2[]
     h > 0 || (h = default_size[1])
     w > 0 || (w = default_size[2])
@@ -500,6 +509,7 @@ function notify_filled(buffer::IOBuffer, nread::Int)
     else
         buffer.ptr += nread
     end
+    nothing
 end
 
 function alloc_buf_hook(stream::LibuvStream, size::UInt)
@@ -539,29 +549,35 @@ function uv_readcb(handle::Ptr{Cvoid}, nread::Cssize_t, buf::Ptr{Cvoid})
     stream_unknown_type = @handle_as handle LibuvStream
     nrequested = ccall(:jl_uv_buf_len, Csize_t, (Ptr{Cvoid},), buf)
     function readcb_specialized(stream::LibuvStream, nread::Int, nrequested::UInt)
-        if nread < 0
-            if nread == UV_ENOBUFS && nrequested == 0
-                # remind the client that stream.buffer is full
-                notify(stream.readnotify)
-            elseif nread == UV_EOF
-                if isa(stream, TTY)
-                    stream.status = StatusEOF # libuv called uv_stop_reading already
-                    notify(stream.readnotify)
-                    notify(stream.closenotify)
-                elseif stream.status != StatusClosing
-                    # begin shutdown of the stream
-                    ccall(:jl_close_uv, Cvoid, (Ptr{Cvoid},), stream.handle)
+        lock(stream.cond)
+        try
+            if nread < 0
+                if nread == UV_ENOBUFS && nrequested == 0
+                    # remind the client that stream.buffer is full
+                    notify(stream.cond)
+                elseif nread == UV_EOF
+                    if isa(stream, TTY)
+                        stream.status = StatusEOF # libuv called uv_stop_reading already
+                        notify(stream.cond)
+                    elseif stream.status != StatusClosing
+                        # begin shutdown of the stream
+                        ccall(:jl_close_uv, Cvoid, (Ptr{Cvoid},), stream.handle)
+                        stream.status = StatusClosing
+                    end
+                else
+                    stream.readerror = _UVError("read", nread)
+                    # This is a fatal connection error. Shutdown requests as per the usual
+                    # close function won't work and libuv will fail with an assertion failure
+                    ccall(:jl_forceclose_uv, Cvoid, (Ptr{Cvoid},), stream)
                     stream.status = StatusClosing
+                    notify(stream.cond)
                 end
             else
-                # This is a fatal connection error. Shutdown requests as per the usual
-                # close function won't work and libuv will fail with an assertion failure
-                ccall(:jl_forceclose_uv, Cvoid, (Ptr{Cvoid},), stream)
-                notify_error(stream.readnotify, _UVError("read", nread))
+                notify_filled(stream.buffer, nread)
+                notify(stream.cond)
             end
-        else
-            notify_filled(stream.buffer, nread)
-            notify(stream.readnotify)
+        finally
+            unlock(stream.cond)
         end
 
         # Stop background reading when
@@ -582,19 +598,24 @@ function uv_readcb(handle::Ptr{Cvoid}, nread::Cssize_t, buf::Ptr{Cvoid})
 end
 
 function reseteof(x::TTY)
+    iolock_begin()
     if x.status == StatusEOF
         x.status = StatusOpen
     end
+    iolock_end()
     nothing
 end
 
 function _uv_hook_close(uv::Union{LibuvStream, LibuvServer})
-    uv.handle = C_NULL
-    uv.status = StatusClosed
-    # notify any listeners that exist on this libuv stream type
-    notify(uv.closenotify)
-    isdefined(uv, :readnotify) && notify(uv.readnotify)
-    isdefined(uv, :connectnotify) && notify(uv.connectnotify)
+    lock(uv.cond)
+    try
+        uv.handle = C_NULL
+        uv.status = StatusClosed
+        # notify any listeners that exist on this libuv stream type
+        notify(uv.cond)
+    finally
+        unlock(uv.cond)
+    end
     nothing
 end
 
@@ -647,12 +668,14 @@ show(io::IO, stream::Pipe) = print(io,
 ## Functions for PipeEndpoint and PipeServer ##
 
 function open_pipe!(p::PipeEndpoint, handle::OS_HANDLE)
+    iolock_begin()
     if p.status != StatusInit
         error("pipe is already in use or has been closed")
     end
     err = ccall(:uv_pipe_open, Int32, (Ptr{Cvoid}, OS_HANDLE), p.handle, handle)
     uv_error("pipe_open", err)
     p.status = StatusOpen
+    iolock_end()
     return p
 end
 
@@ -667,13 +690,11 @@ function link_pipe!(read_end::PipeEndpoint, reader_supports_async::Bool,
             close_pipe_sync(rd)
             rethrow()
         end
-        read_end.status = StatusOpen
         open_pipe!(write_end, wr)
     catch
         close_pipe_sync(wr)
         rethrow()
     end
-    write_end.status = StatusOpen
     nothing
 end
 
@@ -705,6 +726,7 @@ end
 # flow control
 
 function start_reading(stream::LibuvStream)
+    iolock_begin()
     if stream.status == StatusOpen
         if !isreadable(stream)
             error("tried to read a stream that is not readable")
@@ -714,15 +736,16 @@ function start_reading(stream::LibuvStream)
         stream.status = StatusActive
         ret = ccall(:uv_read_start, Cint, (Ptr{Cvoid}, Ptr{Cvoid}, Ptr{Cvoid}),
                     stream, uv_jl_alloc_buf::Ptr{Cvoid}, uv_jl_readcb::Ptr{Cvoid})
-        return ret
     elseif stream.status == StatusPaused
         stream.status = StatusActive
-        return Int32(0)
+        ret = Int32(0)
     elseif stream.status == StatusActive
-        return Int32(0)
+        ret = Int32(0)
     else
-        return Int32(-1)
+        ret = Int32(-1)
     end
+    iolock_end()
+    return ret
 end
 
 if Sys.iswindows()
@@ -732,17 +755,21 @@ if Sys.iswindows()
     # and a ReadFile call blocking on one thread
     # causes all other operations on that stream to lockup
     function stop_reading(stream::LibuvStream)
+        iolock_begin()
         if stream.status == StatusActive
             stream.status = StatusOpen
             ccall(:uv_read_stop, Cint, (Ptr{Cvoid},), stream)
         end
+        iolock_end()
         nothing
     end
 else
     function stop_reading(stream::LibuvStream)
+        iolock_begin()
         if stream.status == StatusActive
             stream.status = StatusPaused
         end
+        iolock_end()
         nothing
     end
 end
@@ -751,107 +778,166 @@ end
 
 readbytes!(s::LibuvStream, a::Vector{UInt8}, nb = length(a)) = readbytes!(s, a, Int(nb))
 function readbytes!(s::LibuvStream, a::Vector{UInt8}, nb::Int)
+    iolock_begin()
     sbuf = s.buffer
     @assert sbuf.seekable == false
     @assert sbuf.maxsize >= nb
 
-    if bytesavailable(sbuf) >= nb
-        return readbytes!(sbuf, a, nb)
+    function wait_locked(s, buf, nb)
+        while bytesavailable(buf) < nb
+            s.readerror === nothing || throw(s.readerror)
+            isopen(s) || break
+            iolock_end()
+            wait_readnb(s, nb)
+            iolock_begin()
+        end
     end
 
     if nb <= SZ_UNBUFFERED_IO # Under this limit we are OK with copying the array from the stream's buffer
-        wait_readnb(s, nb)
-        return readbytes!(sbuf, a, nb)
+        wait_locked(s, sbuf, nb)
+    end
+    if bytesavailable(sbuf) >= nb
+        nread = readbytes!(sbuf, a, nb)
     else
-        try
-            stop_reading(s) # Just playing it safe, since we are going to switch buffers.
-            newbuf = PipeBuffer(a, maxsize = nb)
-            newbuf.size = 0 # reset the write pointer to the beginning
+        newbuf = PipeBuffer(a, maxsize=nb)
+        newbuf.size = 0 # reset the write pointer to the beginning
+        nread = try
             s.buffer = newbuf
             write(newbuf, sbuf)
-            wait_readnb(s, Int(nb))
-            compact(newbuf)
-            return bytesavailable(newbuf)
+            wait_locked(s, newbuf, nb)
+            bytesavailable(newbuf)
         finally
             s.buffer = sbuf
-            if !isempty(s.readnotify.waitq)
-                start_reading(s) # resume reading iff there are currently other read clients of the stream
-            end
         end
+        compact(newbuf)
     end
-    @assert false # unreachable
+    iolock_end()
+    return nread
 end
 
 function read(stream::LibuvStream)
     wait_readnb(stream, typemax(Int))
-    return take!(stream.buffer)
+    iolock_begin()
+    bytes = take!(stream.buffer)
+    iolock_end()
+    return bytes
 end
 
 function unsafe_read(s::LibuvStream, p::Ptr{UInt8}, nb::UInt)
+    iolock_begin()
     sbuf = s.buffer
     @assert sbuf.seekable == false
     @assert sbuf.maxsize >= nb
 
-    if bytesavailable(sbuf) >= nb
-        return unsafe_read(sbuf, p, nb)
+    function wait_locked(s, buf, nb)
+        while bytesavailable(buf) < nb
+            s.readerror === nothing || throw(s.readerror)
+            isopen(s) || throw(EOFError())
+            iolock_end()
+            wait_readnb(s, nb)
+            iolock_begin()
+        end
     end
 
     if nb <= SZ_UNBUFFERED_IO # Under this limit we are OK with copying the array from the stream's buffer
-        wait_readnb(s, Int(nb))
+        wait_locked(s, sbuf, Int(nb))
+    end
+    if bytesavailable(sbuf) >= nb
         unsafe_read(sbuf, p, nb)
     else
+        newbuf = PipeBuffer(unsafe_wrap(Array, p, nb), maxsize=Int(nb))
+        newbuf.size = 0 # reset the write pointer to the beginning
         try
-            stop_reading(s) # Just playing it safe, since we are going to switch buffers.
-            newbuf = PipeBuffer(unsafe_wrap(Array, p, nb), maxsize = Int(nb))
-            newbuf.size = 0 # reset the write pointer to the beginning
             s.buffer = newbuf
             write(newbuf, sbuf)
-            wait_readnb(s, Int(nb))
-            nb == bytesavailable(newbuf) || throw(EOFError())
+            wait_locked(s, newbuf, Int(nb))
         finally
             s.buffer = sbuf
-            if !isempty(s.readnotify.waitq)
-                start_reading(s) # resume reading iff there are currently other read clients of the stream
-            end
         end
     end
+    iolock_end()
     nothing
 end
 
 function read(this::LibuvStream, ::Type{UInt8})
-    wait_readnb(this, 1)
-    buf = this.buffer
-    @assert buf.seekable == false
-    return read(buf, UInt8)
+    iolock_begin()
+    sbuf = this.buffer
+    @assert sbuf.seekable == false
+    while bytesavailable(sbuf) < 1
+        iolock_end()
+        eof(this) && throw(EOFError())
+        iolock_begin()
+    end
+    c = read(sbuf, UInt8)
+    iolock_end()
+    return c
 end
 
 function readavailable(this::LibuvStream)
-    wait_readnb(this, 1)
+    wait_readnb(this, 1) # unlike the other `read` family of functions, this one doesn't guarantee error reporting
+    iolock_begin()
     buf = this.buffer
     @assert buf.seekable == false
-    return take!(buf)
+    bytes = take!(buf)
+    iolock_end()
+    return bytes
 end
 
-function readuntil(this::LibuvStream, c::UInt8; keep::Bool=false)
-    wait_readbyte(this, c)
-    buf = this.buffer
+function readuntil(x::LibuvStream, c::UInt8; keep::Bool=false)
+    iolock_begin()
+    buf = x.buffer
     @assert buf.seekable == false
-    return readuntil(buf, c, keep=keep)
+    if !occursin(c, buf) # fast path checks first
+        x.readerror === nothing || throw(x.readerror)
+        if isopen(x)
+            preserve_handle(x)
+            lock(x.cond)
+            try
+                while !occursin(c, x.buffer)
+                    x.readerror === nothing || throw(x.readerror)
+                    isopen(x) || break
+                    start_reading(x) # ensure we are reading
+                    iolock_end()
+                    wait(x.cond)
+                    unlock(x.cond)
+                    iolock_begin()
+                    lock(x.cond)
+                end
+            finally
+                if isempty(x.cond)
+                    stop_reading(x) # stop reading iff there are currently no other read clients of the stream
+                end
+                unlock(x.cond)
+                unpreserve_handle(x)
+            end
+        end
+    end
+    bytes = readuntil(buf, c, keep=keep)
+    iolock_end()
+    return bytes
 end
 
-uv_write(s::LibuvStream, p::Vector{UInt8}) = uv_write(s, pointer(p), UInt(sizeof(p)))
+uv_write(s::LibuvStream, p::Vector{UInt8}) = GC.@preserve p uv_write(s, pointer(p), UInt(sizeof(p)))
 
+# caller must have acquired the iolock
 function uv_write(s::LibuvStream, p::Ptr{UInt8}, n::UInt)
     uvw = uv_write_async(s, p, n)
     ct = current_task()
     preserve_handle(ct)
-    try
+    sigatomic_begin()
+    uv_req_set_data(uvw, ct)
+    iolock_end()
+    status = try
+        sigatomic_end()
         # wait for the last chunk to complete (or error)
         # assume that any errors would be sticky,
         # (so we don't need to monitor the error status of the intermediate writes)
-        uv_req_set_data(uvw, ct)
-        wait()
+        wait()::Cint
     finally
+        # try-finally unwinds the sigatomic level, so need to repeat sigatomic_end
+        sigatomic_end()
+        iolock_begin()
+        ct.queue === nothing || list_deletefirst!(ct.queue, ct)
         if uv_req_data(uvw) != C_NULL
             # uvw is still alive,
             # so make sure we won't get spurious notifications later
@@ -860,20 +946,24 @@ function uv_write(s::LibuvStream, p::Ptr{UInt8}, n::UInt)
             # done with uvw
             Libc.free(uvw)
         end
+        iolock_end()
         unpreserve_handle(ct)
     end
+    if status < 0
+        throw(_UVError("write", status))
+    end
     return Int(n)
 end
 
 # helper function for uv_write that returns the uv_write_t struct for the write
-# rather than waiting on it
+# rather than waiting on it, caller must hold the iolock
 function uv_write_async(s::LibuvStream, p::Ptr{UInt8}, n::UInt)
     check_open(s)
     while true
         uvw = Libc.malloc(_sizeof_uv_write)
         uv_req_set_data(uvw, C_NULL) # in case we get interrupted before arriving at the wait call
         nwrite = min(n, MAX_OS_WRITE) # split up the write into chunks the OS can handle.
-        # TODO: use writev, when that is added to uv-win
+        # TODO: use writev instead of a loop
         err = ccall(:jl_uv_write,
                     Int32,
                     (Ptr{Cvoid}, Ptr{Cvoid}, UInt, Ptr{Cvoid}, Ptr{Cvoid}),
@@ -897,30 +987,32 @@ end
 # - large isbits arrays are unbuffered and written directly
 
 function unsafe_write(s::LibuvStream, p::Ptr{UInt8}, n::UInt)
-    if s.sendbuf === nothing
-        return uv_write(s, p, UInt(n))
-    end
-
-    buf = s.sendbuf
-    totb = bytesavailable(buf) + n
-    if totb < buf.maxsize
-        nb = unsafe_write(buf, p, n)
-    else
-        flush(s)
-        if n > buf.maxsize
-            nb = uv_write(s, p, n)
-        else
+    while true
+        # try to add to the send buffer
+        iolock_begin()
+        buf = s.sendbuf
+        buf === nothing && break
+        totb = bytesavailable(buf) + n
+        if totb < buf.maxsize
             nb = unsafe_write(buf, p, n)
+            iolock_end()
+            return nb
         end
+        bytesavailable(buf) == 0 && break
+        # perform flush(s)
+        arr = take!(buf)
+        uv_write(s, arr)
     end
-    return nb
+    # perform the output to the kernel
+    return uv_write(s, p, n)
 end
 
 function flush(s::LibuvStream)
+    iolock_begin()
     buf = s.sendbuf
     if buf !== nothing
         if bytesavailable(buf) > 0
-            arr = take!(buf)        # Array of UInt8s
+            arr = take!(buf)
             uv_write(s, arr)
             return
         end
@@ -929,16 +1021,26 @@ function flush(s::LibuvStream)
     return
 end
 
-buffer_writes(s::LibuvStream, bufsize) = (s.sendbuf=PipeBuffer(bufsize); s)
+function buffer_writes(s::LibuvStream, bufsize)
+    sendbuf = PipeBuffer(bufsize)
+    iolock_begin()
+    s.sendbuf = sendbuf
+    iolock_end()
+    return s
+end
 
 ## low-level calls to libuv ##
 
 function write(s::LibuvStream, b::UInt8)
     buf = s.sendbuf
     if buf !== nothing
+        iolock_begin()
         if bytesavailable(buf) + 1 < buf.maxsize
-            return write(buf, b)
+            n = write(buf, b)
+            iolock_end()
+            return n
         end
+        iolock_end()
     end
     return write(s, Ref{UInt8}(b))
 end
@@ -948,12 +1050,7 @@ function uv_writecb_task(req::Ptr{Cvoid}, status::Cint)
     if d != C_NULL
         uv_req_set_data(req, C_NULL) # let the Task know we got the writecb
         t = unsafe_pointer_to_objref(d)::Task
-        if status < 0
-            err = _UVError("write", status)
-            schedule(t, err, error=true)
-        else
-            schedule(t)
-        end
+        schedule(t, status)
     else
         # no owner for this req, safe to just free it
         Libc.free(req)
@@ -1109,61 +1206,99 @@ end
 # BufferStream's are non-OS streams, backed by a regular IOBuffer
 mutable struct BufferStream <: LibuvStream
     buffer::IOBuffer
-    r_c::Condition
-    close_c::Condition
+    cond::Threads.Condition
     is_open::Bool
     buffer_writes::Bool
-    lock::ReentrantLock
+    lock::ReentrantLock # advisory lock
 
-    BufferStream() = new(PipeBuffer(), Condition(), Condition(), true, false, ReentrantLock())
+    BufferStream() = new(PipeBuffer(), Threads.Condition(), true, false, ReentrantLock())
 end
 
 isopen(s::BufferStream) = s.is_open
+
 function close(s::BufferStream)
-    s.is_open = false
-    notify(s.r_c)
-    notify(s.close_c)
-    nothing
+    lock(s.cond) do
+        s.is_open = false
+        notify(s.cond)
+        nothing
+    end
 end
 uvfinalize(s::BufferStream) = nothing
 
-read(s::BufferStream, ::Type{UInt8}) = (wait_readnb(s, 1); read(s.buffer, UInt8))
-unsafe_read(s::BufferStream, a::Ptr{UInt8}, nb::UInt) = (wait_readnb(s, Int(nb)); unsafe_read(s.buffer, a, nb))
+function read(s::BufferStream, ::Type{UInt8})
+    nread = lock(s.cond) do
+        wait_readnb(s, 1)
+        read(s.buffer, UInt8)
+    end
+    return nread
+end
+function unsafe_read(s::BufferStream, a::Ptr{UInt8}, nb::UInt)
+    lock(s.cond) do
+        wait_readnb(s, Int(nb))
+        unsafe_read(s.buffer, a, nb)
+        nothing
+    end
+end
 bytesavailable(s::BufferStream) = bytesavailable(s.buffer)
 
 isreadable(s::BufferStream) = s.buffer.readable
 iswritable(s::BufferStream) = s.buffer.writable
 
 function wait_readnb(s::BufferStream, nb::Int)
-    while isopen(s) && bytesavailable(s.buffer) < nb
-        wait(s.r_c)
+    lock(s.cond) do
+        while isopen(s) && bytesavailable(s.buffer) < nb
+            wait(s.cond)
+        end
     end
 end
 
-show(io::IO, s::BufferStream) = print(io,"BufferStream() bytes waiting:",bytesavailable(s.buffer),", isopen:", s.is_open)
+show(io::IO, s::BufferStream) = print(io, "BufferStream() bytes waiting:", bytesavailable(s.buffer), ", isopen:", s.is_open)
 
-function wait_readbyte(s::BufferStream, c::UInt8)
-    while isopen(s) && !occursin(c, s.buffer)
-        wait(s.r_c)
+function readuntil(s::BufferStream, c::UInt8; keep::Bool=false)
+    bytes = lock(s.cond) do
+        while isopen(s) && !occursin(c, s.buffer)
+            wait(s.cond)
+        end
+        readuntil(s.buffer, c, keep=keep)
+    end
+    return bytes
+end
+
+function wait_close(s::BufferStream)
+    lock(s.cond) do
+        while isopen(s)
+            wait(s.cond)
+        end
     end
 end
 
-wait_close(s::BufferStream) = if isopen(s); wait(s.close_c); end
 start_reading(s::BufferStream) = Int32(0)
 stop_reading(s::BufferStream) = nothing
 
 write(s::BufferStream, b::UInt8) = write(s, Ref{UInt8}(b))
 function unsafe_write(s::BufferStream, p::Ptr{UInt8}, nb::UInt)
-    rv = unsafe_write(s.buffer, p, nb)
-    !(s.buffer_writes) && notify(s.r_c)
-    return rv
+    nwrite = lock(s.cond) do
+        rv = unsafe_write(s.buffer, p, nb)
+        s.buffer_writes || notify(s.cond)
+        rv
+    end
+    return nwrite
 end
 
 function eof(s::BufferStream)
-    wait_readnb(s, 1)
-    return !isopen(s) && bytesavailable(s)<=0
+    bytesavailable(s) > 0 && return false
+    iseof = lock(s.cond) do
+        wait_readnb(s, 1)
+        return !isopen(s) && bytesavailable(s) <= 0
+    end
+    return iseof
 end
 
 # If buffer_writes is called, it will delay notifying waiters till a flush is called.
-buffer_writes(s::BufferStream, bufsize=0) = (s.buffer_writes=true; s)
-flush(s::BufferStream) = (notify(s.r_c); nothing)
+buffer_writes(s::BufferStream, bufsize=0) = (s.buffer_writes = true; s)
+function flush(s::BufferStream)
+    lock(s.cond) do
+        notify(s.cond)
+        nothing
+    end
+end
