@@ -17,25 +17,26 @@ extern "C" {
 
 static jl_sym_t *symtab = NULL;
 
-static uintptr_t hash_symbol(const char *str, size_t len)
+static uintptr_t hash_symbol(const char *str, size_t len) JL_NOTSAFEPOINT
 {
     return memhash(str, len) ^ ~(uintptr_t)0/3*2;
 }
 
-static size_t symbol_nbytes(size_t len)
+static size_t symbol_nbytes(size_t len) JL_NOTSAFEPOINT
 {
     return (sizeof(jl_taggedvalue_t) + sizeof(jl_sym_t) + len + 1 + 7) & -8;
 }
 
-static jl_sym_t *mk_symbol(const char *str, size_t len)
+static jl_sym_t *mk_symbol(const char *str, size_t len) JL_NOTSAFEPOINT
 {
     jl_sym_t *sym;
     size_t nb = symbol_nbytes(len);
+    assert(jl_symbol_type && "not initialized");
 
     jl_taggedvalue_t *tag = (jl_taggedvalue_t*)jl_gc_perm_alloc_nolock(nb, 0, sizeof(void*), 0);
     sym = (jl_sym_t*)jl_valueof(tag);
     // set to old marked so that we won't look at it in the GC or write barrier.
-    tag->header = ((uintptr_t)jl_sym_type) | GC_OLD_MARKED;
+    tag->header = ((uintptr_t)jl_symbol_type) | GC_OLD_MARKED;
     sym->left = sym->right = NULL;
     sym->hash = hash_symbol(str, len);
     memcpy(jl_symbol_name(sym), str, len);
@@ -43,7 +44,7 @@ static jl_sym_t *mk_symbol(const char *str, size_t len)
     return sym;
 }
 
-static jl_sym_t *symtab_lookup(jl_sym_t **ptree, const char *str, size_t len, jl_sym_t ***slot)
+static jl_sym_t *symtab_lookup(jl_sym_t **ptree, const char *str, size_t len, jl_sym_t ***slot) JL_NOTSAFEPOINT
 {
     jl_sym_t *node = jl_atomic_load_acquire(ptree); // consume
     uintptr_t h = hash_symbol(str, len);
@@ -70,7 +71,7 @@ static jl_sym_t *symtab_lookup(jl_sym_t **ptree, const char *str, size_t len, jl
     return node;
 }
 
-static jl_sym_t *_jl_symbol(const char *str, size_t len)
+static jl_sym_t *_jl_symbol(const char *str, size_t len) JL_NOTSAFEPOINT
 {
     jl_sym_t **slot;
     jl_sym_t *node = symtab_lookup(&symtab, str, len, &slot);
@@ -118,9 +119,9 @@ JL_DLLEXPORT jl_sym_t *jl_gensym(void)
 {
     char name[16];
     char *n;
-    n = uint2str(&name[2], sizeof(name)-2, gs_ctr, 10);
+    uint32_t ctr = jl_atomic_fetch_add(&gs_ctr, 1);
+    n = uint2str(&name[2], sizeof(name)-2, ctr, 10);
     *(--n) = '#'; *(--n) = '#';
-    gs_ctr++;
     return jl_symbol(n);
 }
 
@@ -134,9 +135,9 @@ JL_DLLEXPORT jl_sym_t *jl_tagged_gensym(const char *str, int32_t len)
     char *n;
     name[0] = '#'; name[1] = '#'; name[2+len] = '#';
     memcpy(name+2, str, len);
-    n = uint2str(gs_name, sizeof(gs_name), gs_ctr, 10);
+    uint32_t ctr = jl_atomic_fetch_add(&gs_ctr, 1);
+    n = uint2str(gs_name, sizeof(gs_name), ctr, 10);
     memcpy(name+3+len, n, sizeof(gs_name)-(n-gs_name));
-    gs_ctr++;
     jl_sym_t *sym = _jl_symbol(name, len+3+sizeof(gs_name)-(n-gs_name)-1);
     if (len >= 256) free(name);
     return sym;
